package org.example.newmat;
import Graph.readFile;

import javafx.animation.AnimationTimer;
import javafx.application.Application;
import javafx.geometry.Bounds;
import javafx.geometry.Pos;
import javafx.scene.*;
import javafx.scene.control.Label;
import javafx.scene.input.KeyEvent;
import javafx.scene.layout.StackPane;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.*;
import javafx.scene.transform.Rotate;
import javafx.scene.transform.Translate;
import javafx.stage.Stage;

import java.io.IOException;
import java.util.Arrays;


/**
 * Demonstrates advanced JavaFX graphics and window management by creating a 3D scene with
 * animated elements and FPS (Frames Per Second) counter.
 * <p>
 * This application creates a 3D scene containing a textured cube and a dynamically generated
 * mesh based on input parameters. It showcases the handling of 3D transformations, animation,
 * camera movement, and basic input from the keyboard to manipulate the scene. Additionally,
 * it includes a real-time FPS counter overlaid on the scene to monitor performance.
 */
public class GraphicsAndWindowsTest extends Application {
    public static String[] parameters;

    /**
     * Starts the JavaFX application.
     * This method initializes the 3D scene with its objects, animation timer, and event handlers.
     *
     * @param stage The primary stage for this application, onto which the scene is set.
     * @throws IOException If there is an issue reading the input file for generating the mesh.
     */
    @Override
    public void start(Stage stage) throws IOException {
        Box cube = new Box(300,300,300);
        cube.setDrawMode(DrawMode.LINE);
        Translate translate1 = new Translate();
        // Set the translation values
        //Cube is centered at (600,300,150)
        translate1.setX(750);
        translate1.setY(450);
        translate1.setZ(150);

        cube.getTransforms().add(translate1);
        TriangleMesh mesh = new TriangleMesh();

        /**
         * Texture coordinates used to add image to a mesh.
         * Initialized at 0, 0 since there is no need for an image.
         */
        float[] texCoords = {
                0, 0
        };

        parameters = getParameters().getRaw().toArray(new String[getParameters().getRaw().size()]);
        readFile.generateRandomVertices(1000);
        readFile.randFileGenerator("testfile2.txt");
        readFile processInput = new readFile();
        processInput.parseCommand(parameters);
        processInput.makeVertexHash(processInput.path[0]);
        //processInput.makeMatrix();
//        System.out.print(processInput.adjMatrix.toString());
//        System.out.print(Arrays.toString(processInput.getVertices()));
//        System.out.print(Arrays.toString(processInput.getFaces()));
        //Gives the vertices (points) to the mesh in the form of a float [] array
        mesh.getPoints().setAll(processInput.getVertices());

        //Gives faces to the mesh in the form of an int [] array
        mesh.getFaces().setAll(processInput.getFaces());

        //gives texture coordinate array
        mesh.getTexCoords().setAll(texCoords);

        // This is where the mesh is set to the MeshView.
        MeshView triangleView = new MeshView(mesh);
        // Removes front and back differential
        // removes need to keep track of front and back
        triangleView.setCullFace(CullFace.NONE);
        triangleView.setMaterial(new PhongMaterial(Color.BLUE));

        // Create a Translate transform
        Translate translate = new Translate();
        // Set the translation values
        translate.setX(600); // Translate 600 units along the X axis
        translate.setY(300); // Translate 300 units along the Y axis
        translate.setZ(0);   // Translate 0 units along the Z axis

        triangleView.getTransforms().add(translate);

        Group group3D = new Group(); // Group of our 3D elements
        group3D.getChildren().add(triangleView);
        group3D.getChildren().add(cube);
        group3D.getChildren().add(new AmbientLight(Color.WHITE));

        // NOTE: Due to the perspectiveCamera it looks like it is rotating even though it isn't
        // If you comment out this line and the scene.setCamera(camera) line
        // You will see it move without the rotation effect
        // I think this is pretty cool
        Camera camera = new PerspectiveCamera();
        camera.setNearClip(.1);
        camera.setFarClip(2000);
        camera.setTranslateX(0);
        camera.setTranslateY(0);
        camera.setTranslateZ(300);

        //https://stackoverflow.com/questions/65510979/move-camera-relative-to-view-in-javafx
        //Incorporated the code from this post
        Rotate groupRotX = new Rotate(0, Rotate.X_AXIS);
        Rotate groupRotY = new Rotate(0, Rotate.Y_AXIS);


        group3D.getTransforms().addAll(groupRotY, groupRotX);

        stage.addEventHandler(KeyEvent.KEY_PRESSED, event -> {
            switch(event.getCode()){
                case LEFT:
                    groupRotY.setAngle(groupRotY.getAngle() + 10);
                    break;
                case RIGHT:
                    groupRotY.setAngle(groupRotY.getAngle() - 10);
                    break;
                case UP:
                    groupRotX.setAngle(groupRotX.getAngle() + 10);
                    break;
                case DOWN:
                    groupRotX.setAngle(groupRotX.getAngle() - 10);
                    break;

            }
        });

        /*
         * SubScene to handle 3D components
         * Makes it easier to have 2D and 3D components in the same animation
         */
        SubScene subScene3D = new SubScene(group3D, 1500, 800);
        subScene3D.setFill(Color.SILVER);
        subScene3D.setCamera(camera);



        /*
         * Text label to display FPS
         */
        Label fpsLabel = new Label("FPS: 0");
        fpsLabel.setTextFill(Color.WHITE);

        //Lays out the children in Z order with the last member being front most
        //Effectively allows you to layer multiple elements
        //We are layering our 3D scene with a 2D label on top of it
        StackPane Layers = new StackPane();
        Layers.getChildren().addAll(subScene3D, fpsLabel);
        StackPane.setAlignment(fpsLabel, Pos.BOTTOM_RIGHT); //Places FPS in bottom right

        Scene scene = new Scene(Layers, 1500, 800);

        stage.setTitle("FPS test");
        stage.setScene(scene);
        stage.show();

        // Animation logic to move the Box back and forth
        AnimationTimer timer = new AnimationTimer() {
            private long lastUpdate = 0; //Movement timer
            private long fpsTimer = 0; //Frame timer
            private long frameRate = 0; //Frames per second
            private double directionY = 5.0; // Direction of movement
            private double directionX = 4.0; // Direction of movement
            private double directionZ = 10.0; // Direction of movement

            /**
             * Called at every frame while the {@link AnimationTimer} is active to update the scene's elements,
             * including object movements, boundary condition checks, mesh deformation on collision, and FPS counter updates.
             * <p>
             * This method extends the scene update logic with collision detection between the mesh and a virtual bounding box,
             * causing mesh deformation upon nearing the box's edges. It calculates new positions for 3D objects based on their
             * current positions and direction vectors, simulates movement within the scene, and reverses the movement direction
             * at the scene's boundaries. Additionally, it modifies the mesh's vertices to "squish" the mesh when it comes close
             * to or collides with the bounding box, creating a more dynamic interaction between the scene's objects.
             * <p>
             * Movement, boundary checking, and mesh deformation logic are applied based on the bounds of the 3D mesh and cube within the scene.
             * The FPS counter is updated once every second, based on the number of frames rendered in that interval, to provide real-time
             * performance feedback.
             *
             * @param now The timestamp of the current frame given in nanoseconds. This value is used to ensure smooth animations and
             *            consistent updates across frames.
             */
            @Override
            public void handle(long now) {
                //Built in class that helps calculate the bounds of the mesh for collision
                Bounds meshBounds = triangleView.getBoundsInParent();

                double minX = meshBounds.getMinX() + directionX;
                double maxX = meshBounds.getMaxX() + directionX;
                double minY = meshBounds.getMinY() + directionY;
                double maxY = meshBounds.getMaxY() + directionY;
                double minZ = meshBounds.getMinZ() + directionZ;
                double maxZ = meshBounds.getMaxZ() + directionZ;
                //Number is roughly 1 mill divided by 60
                if (now - lastUpdate >= 16_000_000) { // Roughly 60 frames per second

                    double cubeMin = 300, cubeMax = 600;
                    // Reverse direction at bounds
                    if (minX < cubeMin+300 || maxX > cubeMax+300) {

                        // load the vertices of the mesh
                        float[] points = mesh.getPoints().toArray(null);

                        // find the vertices which come close to the bounding box threshold and squish them into the mesh
                        for (int j = 0; j < points.length; j += 3) {
                            if (Math.abs(points[j] + translate.getX() + directionX - minX) < 10 || Math.abs(points[j] + translate.getX() + directionX - maxX) < 10) {
                                points[j] += -directionX*5;
                            }
                        }

                        //update the direction of the mesh and set the new coordinates
                        directionX *= -1;
                        mesh.getPoints().setAll(points);
                    }
                    if (minY < cubeMin || maxY > cubeMax) {

                        // load the vertices of the mesh
                        float[] points = mesh.getPoints().toArray(null);

                        // find the vertices which come close to the bounding box threshold and squish them into the mesh
                        for (int j = 1; j < points.length; j += 3) {
                            if (Math.abs(points[j] + translate.getY() + directionY - minY) < 10 || Math.abs(points[j] + translate.getY() + directionY - maxY) < 10) {
                                points[j] += -directionY*5;
                            }
                        }

                        //update the direction of the mesh and set the new coordinates
                        directionY *= -1;
                        mesh.getPoints().setAll(points);
                    }
                    if (minZ < cubeMin-300 || maxZ > cubeMax-300) {

                        // load the vertices of the mesh
                        float[] points = mesh.getPoints().toArray(null);

                        // find the vertices which come close to the bounding box threshold and squish them into the mesh
                        for (int j = 2; j < points.length; j += 3) {
                            if (Math.abs(points[j] + translate.getZ() + directionZ - minZ) < 10 || Math.abs(points[j] + translate.getZ() + directionZ - maxZ) < 10) {
                                points[j] += -directionZ*5;
                            }
                        }

                        //update the direction of the mesh and set the new coordinates
                        directionZ *= -1;
                        mesh.getPoints().setAll(points);

                    }

//                    if (newZ < 0 || newZ >300) {
//                        directionZ = -directionZ;
//                    }

                    double newY = translate.getY() + directionY;
                    translate.setY(newY);
                    double newX = translate.getX() + directionX;
                    translate.setX(newX);
                    double newZ = translate.getZ() + directionZ;
                    translate.setZ(newZ);

                    lastUpdate = now;
                }
                if (fpsTimer == 0) {
                    fpsTimer = now; // Initialize the FPS timer
                }
                frameRate++; // Increment frame count for each invocation
                //1 mill nano seconds in a second
                if (now - fpsTimer >= 1_000_000_000) { // Every second, update the FPS display
                    fpsLabel.setText(String.format("FPS: %d", frameRate));
                    frameRate = 0; // Reset frame count
                    fpsTimer = now; // Reset the FPS timer
                }
            }
        };
        timer.start();
    }
    /**
     * The entry point of the application. This method is called when the application starts.
     *
     * @param args Command-line arguments passed to the application. Expected to contain
     *             parameters for mesh generation and other configurations.
     */
    public static void main(String[] args) {
        Application.launch(args);
    }
}
